#! /usr/bin/perl -w
#
# $Id: fwd_nagios_results.pl 235 2006-10-05 16:40:27Z mibe $
#

#
# for usage please have a lok at function usage() or call this progamm 
#   ./fwd_nagios_results.pl -h
#

#-------------------------------------------------------------------------------
# configuration and definitions
#-------------------------------------------------------------------------------

use strict;
use sigtrap;
use Getopt::Std;
use Fcntl; # needed for sysopen
use POSIX;
use Sys::Syslog;
use Socket;

$| = 1;

#---------- clear environment

$ENV{'PATH'}='';
$ENV{'BASH_ENV'}='';
$ENV{'ENV'}='';

#---------- global variables
# this is our unknown file descriptor definition
*MY_UNKNOWN_FD = -1;
#
# normal global variables
my $program = "$0";
$program =~ s/.*[\/|\\]//g;
my $PID = $$;
my $PPID = $$;
my %option = (
      'daemon'              => 0,
      'iproto'              => "stdin",
      'isplay'              => 5,
      'filter'              => 0,
      'filter_file'         => "",
      'help'                => 0,
      'listen_queue'        => 5,
      'local_addr'          => "127.0.0.1",
      'local_port'          => 0,
      'oproto'              => "stdout",
      'osplay'              => 0,
      'print'               => 0,
      'quiet'               => 0,
      'remote_addr'         => "127.0.0.1",
      'remote_port'         => 0,
      'reopen_dst'          => 0,
      'tproto'              => "",
      'tee_addr'            => "127.0.0.1",
      'tee_port'            => 0,
      'reopen_tee'          => 0,
      'syslog'              => 0,
      'strip_newline'       => 0
   );
my %default_values = (
      'isplay'              => 5,
      'osplay'              => 1 
   );
my %file_desc = (
      'src'                 => *MY_UNKNOWN_FD,
      'listen'              => *MY_UNKNOWN_FD,
      'dst'                 => *MY_UNKNOWN_FD,
      'tee'                 => *MY_UNKNOWN_FD
   );
my @filter_arr = ();

#-------------------------------------------------------------------------------
#---------- necessary functions
#
#  -c <config file>
#    Location of config file for this program.
#
sub usage {
  print STDERR << "EOF";
usage: progname [<options>]
options:
  -d
    Daemonize this program.
  -f|-F <filter file>
    Use filters from <filter file> as positive (-f) or
    negative (-F) filter. If you use an positive filter
    then only messages that match one of that filter
    will be forwarded. Negative filter filters out
    messages instead.
  -h
    Print this message and exit.
  -i <url>
    Use <url> as source for incomming messages.
    If option -i was omitted this means read from stdin.
  -l
    Write messages to syslog if available.
  -n
    Strip newline from incoming lines.
  -o <url>
    Use <url> as destination for outgoing messages.
    If option -o was omitted this means writo to stdout.
  -p
    Print what you got from input source to stderr.
    This is only for debugging.
  -q
    Be quiet means no info messages.
  -r
    Reopen destination for each outgoing message.
  -R
    Reopen tee destination for each outgoing message.
  -s <int>
    Splay time in seconds to recheck if incomming source
    has appeared.
    (default: $default_values{'isplay'} seconds).
  -t <url>
    Use <url> as tee destination for outgoing messages.
    If option -t was omitted this means no tee output.
    If you want pipe tee output to stdout use "stdout:"
    as url.

As <url> you can specify:
  * named pipes
      npipe:<name of named pipe>
    example:
      npipe:/tmp/incomimg.named_pipe
  * raw tcp streams
      tcp:<hostname or ip address>[:port]
      port is optional
    example:
      tcp:localhost
      tcp:localhost:34567
      tcp:127.0.0.1:4567
      tcp:[::1]:5678     (IPv6 currently not supported)
  * command stream
      cmd:<command line>
    example:
     "cmd:/usr/bin/tail -f /src/file |"
     "cmd:| /usr/bin/tee -a /dst/file"
     "cmd:< /src/file"
     "cmd:> /dst/file"
     "cmd:>> /dst/file"
      
EOF
}

# wrapper to got a clean exit
sub my_exit {
  foreach (sort keys %file_desc) {
    if ($file_desc{$_} ne *MY_UNKNOWN_FD) {
      close($file_desc{$_});
      $file_desc{$_} = *MY_UNKNOWN_FD;
    }
  }

  # close syslog
  if ($option{'syslog'} > 1) {
    closelog();
  }

  _exit($_[0]);
}

# we may need this for filehandles
# so print to an invalid FD does not abort program
sub catch_sigpipe {
  local $!; # don't let someone overwrite current error

  #print(STDERR "sigpipe catched\n");

  # reinstall handler
  $SIG{PIPE} = \&catch_sigpipe;
}
# install handler
$SIG{PIPE} = \&catch_sigpipe;

# in case of we fork something catch our child status
sub catch_sigchild {
  local $!; # don't let waitpid() overwrite current error

  while ((my $pid = waitpid(-1, WNOHANG)) > 0 && WIFEXITED($?)) {
    #print STDERR "catch_sigchild[$PID]: reaped $pid"
    #              . ($? ? " with exit status $?" : '') . "\n";
    next;
  }

  # reinstall handler
  $SIG{CHLD} = \&catch_sigchild;
}
# install handler
$SIG{CHLD} = \&catch_sigchild;

# read our config file
sub read_config {
  return 1;
}

# split url in proto , address and port
# where
#  $_[1] = proto or errmsg
#  $_[2] = name or inet address or command or whatever
#  $_[3] = port (optional)
sub split_url {
  my $peer = $_[0];
  my $proto;
  my $addr;
  my $port = "0";

  if ($peer =~ /^npipe:.*$/) {
    $addr = $peer;
    $addr =~ s/^npipe:(.*)$/$1/;

    $_[1] = "npipe";
    $_[2] = $addr;
  } elsif ($peer =~ /^tcp:.*$/) {
    $addr = $peer;
    $port = $peer;

    if ($peer =~ /^tcp:\[.*$/) {
      $addr =~ s/^tcp:\[(.*)\].*$/$1/;
      $port =~ s/^tcp:\[.*\]:(\d+).*$/$1/;
    } else {
      $addr =~ s/^tcp:(.*[^:]):.*$/$1/;
      $port =~ s/^tcp:.*[^:]:(\d+)$/$1/;
    }

    if ($port =~ /^tcp:/) {
      $port=0;
    }

    $_[1] = "tcp";
    $_[2] = $addr;
    $_[3] = $port;
  } elsif ($peer =~ /^cmd:.*$/) {
    $addr = $peer;
    $addr =~ s/^cmd:(.*)$/$1/;

    $_[1] = "cmd";
    $_[2] = $addr;
  } elsif ($peer =~ /^stdout:.*$/) {
    $_[1] = "stdout";
  } else {
    # not OK
    $addr = $peer;
    $addr =~ s/^([^:]*).*/$1/;
                                                                                                                                                          
    $_[1] = "protocol \"".$addr."\" unknown";
    return 0;
  }
                                                                                                                                                          
  # was OK
  return 1;
}

# got this function from
# http://perldoc.perl.org/perlipc.html
sub daemonize {
  chdir '/'
    or die "error: can't chdir to /: $!";
  open STDIN, '/dev/null'
    or die "error: can't read /dev/null: $!";
  open STDOUT, '>/dev/null'
    or die "error: can't write to /dev/null: $!";
  defined(my $pid = fork)
    or die "error: can't fork: $!";
  if ($pid != 0) {
    my_exit(0);
  }
  POSIX::setsid()
    or die "error: can't start a new session: $!";
  open STDERR, '>&STDOUT'
    or die "error: can't dup stdout: $!";
  $PID = $$;
  $PPID = $$;
}

# to collect file handle bits
sub fhbits {
  my @fhlist = split(' ',$_[0]);
  my $bits = '';
  for (@fhlist) {
    vec($bits,fileno($_),1) = 1;
  }
  $bits;
}

# open a namend pipe source
sub open_source_npipe {

  # if we work with a namend pipe as source
  # check if named pipe is there
  if (! -p $option{'local_addr'}) {
    # try to create our named pipe
    #print STDERR "creating npipe $option{'local_addr'}\n";
    if (!POSIX::mkfifo($option{'local_addr'}, 0666)) {
      if ($option{'syslog'} > 1) {
        syslog('error', "can not create named pipe \"%s\": %s", 
             $option{'local_addr'}, $!);
      } else {
        printf(STDERR "error: can not create named pipe \"%s\": %s\n",
             $option{'local_addr'}, $!);
      }
      # source is not open
      return 1;
    }
  }

  # Check if named pipe is there.
  if ($file_desc{'src'} eq *MY_UNKNOWN_FD) {
    local *SRC_FD;
    if (!sysopen(SRC_FD, $option{'local_addr'}, O_RDONLY)) {
      if ($option{'syslog'} > 1) {
        syslog('error', "can not read from named pipe \"%s\": %s",
               $option{'local_addr'}, $!);
      } else {
        printf(STDERR "error: can not open named pipe \"%s\": %s",
               $option{'local_addr'}, $!);
      }
      $file_desc{'src'} = *MY_UNKNOWN_FD;
      # source is not open
      return 1;
    } else {
      $file_desc{'src'} = *SRC_FD;
    }
  }

  #print STDERR "npipe[$PID]: src is open\n";
  # source is open
  return 0;
}

# forward output into named pipe sink
sub fwd_to_npipe {
  local *DST_FD;
  my $dst = $_[0];
  my $raddr = $_[1];
  my $line = $_[2];

  # open connection to destination if not done already
  if ($file_desc{$dst} eq *MY_UNKNOWN_FD) {
    if (!sysopen(DST_FD, $option{$raddr}, O_WRONLY)) {
      if ($option{'syslog'} > 1) {
        syslog('error', "can not write to \"%s\": %s",
               $option{$raddr}, $!);
      } else {
        printf(STDERR "error: can not write to \"%s\": %s\n",
               $option{$raddr}, $!);
      }
      $file_desc{$dst} = *MY_UNKNOWN_FD;
      # destination is not open
      return 1;
    } else {

      # make communication unbuffered
      my $old_fd = select(DST_FD);
      $| = 1;
      select($old_fd);

      $file_desc{$dst} = *DST_FD;
    }
  }

  # write data
  *DST_FD = $file_desc{$dst};
  if (!print(DST_FD "$line")) {
    print STDERR "npipe[$PID]: print not OK, closing dst \"$dst\"\n";
    close(DST_FD);
    $file_desc{$dst} = *MY_UNKNOWN_FD;
    return 1;
  }
  #print STDERR "npipe[$PID]: $line";

  # if we should reopen destination each time we send a outgoing
  # message, this means we should close it now.
  if ($option{'reopen_'.$dst} == 1) {
    close(DST_FD);
    $file_desc{$dst} = *MY_UNKNOWN_FD;
  }

  return 0;
}

# open a inet listen port
sub open_source_tcp {
  my $pid;
  my $remote_port;
  my $remote_addr;
  my $remote_peer_addr;

  # if not configured
  if ($file_desc{'listen'} eq *MY_UNKNOWN_FD) {
    local *LISTEN_FD;
    my $proto;
    my $local_port;
    my $local_addr;
    my $local_peer_addr;

    # prepare port, address, proto, etc.
    $proto = getprotobyname('tcp');
    $local_port = $option{'local_port'};
    $local_addr = inet_aton($option{'local_addr'});
    $local_peer_addr = pack_sockaddr_in($local_port, $local_addr);

    # create socket where we want listen
    if ($file_desc{'listen'} eq *MY_UNKNOWN_FD) {
      if(!socket(LISTEN_FD, PF_INET, SOCK_STREAM, $proto)) {
        # source is not open
        $file_desc{'listen'} = *MY_UNKNOWN_FD;
        return 1;
      }
      $file_desc{'listen'} = *LISTEN_FD;
    }

    # Set up the port so it's freed as soon as we're done.
    setsockopt(LISTEN_FD, SOL_SOCKET, SO_REUSEADDR, pack("l",1));

    # configure our local peer
    if (!bind($file_desc{'listen'}, $local_peer_addr)) {
      close($file_desc{'listen'});
      $file_desc{'listen'} = *MY_UNKNOWN_FD;
      return 1;
    }
  }

  # wait for next client
  if (!listen($file_desc{'listen'}, $option{'listen_queue'})) {
    # somewhat went wrong so close socket
    print STDERR "tcp[$PID]: listen returns with error\n";
    close($file_desc{'listen'});
    $file_desc{'listen'} = *MY_UNKNOWN_FD;
    return 1;
  }

  # get handle for next connection
  if ($file_desc{'src'} eq *MY_UNKNOWN_FD) {
    local *DST_FD = *MY_UNKNOWN_FD;
    my $old_fd;

    # parent waits on clients on their connect
    while (1) {
      #print STDERR "tcp[$PID]: waiting on accept ...";
      $remote_peer_addr = accept(DST_FD, $file_desc{'listen'});
      if ($remote_peer_addr) {
        #print STDERR "done\n";
        last;
      }
    }

    # make communication unbuffered
    $old_fd = select(DST_FD);
    $| = 1;
    select($old_fd);

    $file_desc{'src'} = *DST_FD;
  }
  ($remote_port, $remote_addr) = unpack_sockaddr_in($remote_peer_addr);

  if ($option{'print'} == 1) {
    if ($option{'syslog'} > 1) {
      syslog('notice', 'client with %s:%d connected', 
             inet_ntoa($remote_addr),
             $remote_port);
    } else {
      printf(STDERR "tcp[$PID]: client with %s:%d connected\n", 
             inet_ntoa($remote_addr),
             $remote_port);
    }
  }

  # check if client could connect

  # client is ok 
  # so fork and child should handle this client
  # parent should go to accept the next
  if (!defined($pid = fork)) {
    # error
    print STDERR "tcp[$PID]: cannot fork: $!";
    return 1;
  } elsif ($pid) {
    # I'm the parent
    #print STDERR "tcp[$PID]: parent goes to accept next client";

    # close source FD and prepare for next client
    close($file_desc{'src'});
    $file_desc{'src'} = *MY_UNKNOWN_FD;

    # flag an not open channel so we come to next accept
    return 1;
  }

  # I'm the child
  $PID = $$;

  #print STDERR "tcp[$PID]: src is open\n";
  # source is open
  return 0;
}

# forward output into one connected inet sink
sub fwd_to_tcp {
  local *DST_FD;
  my $dst = $_[0];
  my $raddr = $_[1];
  my $rport = $_[2];
  my $line = $_[3];

  # open connection to destination if not done already
  if ($file_desc{$dst} eq *MY_UNKNOWN_FD) {
    my $ip_fam;
    my $proto;
    my $remote_port;
    my $remote_addr;
    my $remote_peer_addr;
    my $old_fd;

    # prepare port, address, proto, etc.
    $proto = getprotobyname('tcp');
    $remote_port = $option{$rport};
    if ($option{$raddr} =~ /:/) {
      $remote_addr = ipv6_aton($option{$raddr});
      $ip_fam = PF_INET6;
      $remote_peer_addr = pack_sockaddr_in6($remote_port, $remote_addr);
    } else {
      $remote_addr = inet_aton($option{$raddr});
      $ip_fam = PF_INET;
      $remote_peer_addr = pack_sockaddr_in($remote_port, $remote_addr);
    }

    # create socket where we want transmit our data
    if(!socket(DST_FD, $ip_fam, SOCK_STREAM, $proto)) {
      # destination is not open
      return 1;
    }

    # Set up the port so it's freed as soon as we're done.
    setsockopt(DST_FD, SOL_SOCKET, SO_REUSEADDR, pack("l",1));
 
    # configure our local side of peer if needed
    # if (!bind(DST_FD, $local_peer_addr)) {
    #  # destination is not open
    #  close(DST_FD);
    #  return 1;
    #}

    # connect to our remote peer 
    if (!connect(DST_FD, $remote_peer_addr)) {
      # destination is not open
      close(DST_FD);
      return 1;
    }

    # make communication unbuffered
    $old_fd = select(DST_FD);
    $| = 1;
    select($old_fd);

    # save file descriptor
    $file_desc{$dst} = *DST_FD;
  }

  # write data
  *DST_FD = $file_desc{$dst};
  if (!print(DST_FD "$line")) {
    print STDERR "tcp[$PID]: print not OK, closing dst \"$dst\"\n";
    close(DST_FD);
    $file_desc{$dst} = *MY_UNKNOWN_FD;
    return 1;
  }
  #print STDERR "tcp[$PID]: $line";

  # if we should reopen destination each time we send a outgoing
  # message, this means we should close it now.
  if ($option{'reopen_'.$dst} == 1) {
    close(DST_FD);
    $file_desc{$dst} = *MY_UNKNOWN_FD;
  }

  return 0;
}

# open a pipe from to read from one command
sub open_source_cmd {
  
  # Check if channel is open
  if ($file_desc{'src'} eq *MY_UNKNOWN_FD) {
    local *SRC_FD;
    if (!open(SRC_FD, $option{'local_addr'})) {
      if ($option{'syslog'} > 1) {
        syslog('error', 'can not read from "%s": %s',
               $option{'local_addr'}, $!);
      } else {
        printf(STDERR "error: can not read from \"%s\": %s\n",
               $option{'local_addr'}, $!);
      }
      $file_desc{'src'} = *MY_UNKNOWN_FD;
      # source is not open
      return 1;
    } else {
      $file_desc{'src'} = *SRC_FD;
    }
  }

  #print STDERR "cmd[$PID]: src is open\n";
  # source is open
  return 0;
}

# forward output into a pipe to a command
sub fwd_to_cmd {
  local *DST_FD;
  my $dst = $_[0];
  my $raddr = $_[1];
  my $line = $_[2];

  # open connection to destination if not done already
  if ($file_desc{$dst} eq *MY_UNKNOWN_FD) {
    if (!open(DST_FD, $option{$raddr})) {
      if ($option{'syslog'} > 1) {
        syslog('error', 'can not write to "%s": %s',
               $option{$raddr}, $!);
      } else {
        printf(STDERR "error: can not write to \"%s\": %s\n",
               $option{$raddr}, $!);
      }
      $file_desc{$dst} = *MY_UNKNOWN_FD;
      # destination is not open
      return 1;
    } else {
      $file_desc{$dst} = *DST_FD;
    
      # make communication unbuffered
      my $old_fd = select(DST_FD);
      $| = 1;
      select($old_fd);
    }
  }

  # write data
  *DST_FD = $file_desc{$dst};
  if (!print(DST_FD "$line")) {
    print STDERR "cmd[$PID]: print not OK, closing dst \"$dst\"\n";
    close(DST_FD);
    $file_desc{$dst} = *MY_UNKNOWN_FD;
    return 1;
  }
  #print STDERR "cmd[$PID]: $line";

  # if we should reopen destination each time we send a outgoing
  # message, this means we should close it now.
  if ($option{'reopen_'.$dst} == 1) {
    close(DST_FD);
    $file_desc{$dst} = *MY_UNKNOWN_FD;
  }

  return 0;
}

# init our filter array
sub read_filter {
  local *FILTER_FD;
  my $i;
  my $line;

  if (!open(FILTER_FD, $option{'filter_file'})) {
    if ($option{'syslog'} > 1) {
      syslog('error', 'can not read from filter file "%s": %s',
             $option{'local_addr'}, $!);
    } else {
      printf(STDERR "error: can not read from filter file \"%s\": %s\n",
             $option{'local_addr'}, $!);
    }
    return 1;
  }

  $i = 0; 
  while($line = <FILTER_FD>) {
    chop($line);
    if ($line =~ /^#.*/ ) {
      # comment line
      next;
    } elsif ($line =~ /^$/ ) {
      # empty lines
      next;
    }

    if ($line =~ /^\/(.*)\/\s*$/) {
      # strip slash from begin and end
      $line = $1;
    }

    #all other lines are treated as pattern
    $filter_arr[$i] = qr/$line/;
    $i++;
  }
  close(FILTER_FD); 

  return 0;
}

# did one of our filters match?
sub reject_by_filter {
  my $success;
  my $neg_success;
  my $pat;

  if ($option{'filter'} == 1) {
    # configure for positive filter
    $success = 1;
    $neg_success = 0;
   
  } elsif ($option{'filter'} == 2) {
    # configure for negative filter
    $success = 0;
    $neg_success = 1;
  } else {
    # nothing to do means not rejected
    return 0;
  }

  # scan if pattern matches
  grep {
      foreach $pat (@filter_arr) {
        if (/$pat/) {
          $success = $neg_success;
          last;
        }
      }
      $success;
    } @_;

  return $success;
}

#-------------------------------------------------------------------------------
#---------- process and correct command line options
my %options = ();
getopts("df:F:hi:lno:pqrRs:t:",\%options);

# daemonize
if ($options{'d'}) {
  if ($options{'d'} != 0) {
    $option{'daemon'} = 1;
  }
}

# filter
if ($options{'f'}) {
  if ($options{'f'} !~ /^$/) {
    # positive filter
    $option{'filter'} = 1;
    $option{'filter_file'} = $options{'f'};
  }
} elsif ($options{'F'}) {
  if ($options{'F'} !~ /^$/) {
    # negative filter
    $option{'filter'} = 2;
    $option{'filter_file'} = $options{'F'};
  }
}

# help
if ($options{'h'}) {
  $option{'help'} = 1;
  usage();
  my_exit(0);
}

# source
if ($options{'i'}) {
  if (!split_url($options{'i'}, $option{'iproto'},
                  $option{'local_addr'}, $option{'local_port'})) {
    usage();
    printf(STDERR "error: problem with source url \"%s\".\n", $option{'iproto'});
    my_exit(1);
  }
}

# syslog
if ($options{'l'}) {
  if ($options{'l'} != 0) {
    $option{'syslog'} = $options{'l'};
  } else {
    $option{'syslog'} = 1;
  }
}

# strip newline
if ($options{'n'}) {
  if ($options{'n'} != 0) {
    $option{'strip_newline'} = 1;
  }
}

# destination
if ($options{'o'}) {
  if (!split_url($options{'o'}, $option{'oproto'},
                  $option{'remote_addr'}, $option{'remote_port'})) {
    usage();
    printf(STDERR "error: problem with destination url \"%s\".\n", $option{'oproto'});
    my_exit(1);
  }
}

# print what you got.
if ($options{'p'}) {
  if ($options{'p'} != 0) {
    $option{'print'} = 1;
  }
}

# be quiet
if ($options{'q'}) {
  if ($options{'q'} != 0) {
    $option{'quiet'} = 1;
  }
}

# reopen destionation
if ($options{'r'}) {
  if ($options{'r'} != 0) {
    $option{'reopen_dst'} = 1;
  }
}

# reopen tee destionation
if ($options{'R'}) {
  if ($options{'R'} != 0) {
    $option{'reopen_tee'} = 1;
  }
}

# splaytime
if ($options{'s'}) {
  if ($options{'s'} != 0) {
    $option{'isplay'} = $options{'s'};
  }
}

# tee destination
if ($options{'t'}) {
  if (!split_url($options{'t'}, $option{'tproto'},
                  $option{'tee_addr'}, $option{'tee_port'})) {
    usage();
    printf(STDERR "error: problem with tee destination url \"%s\".\n", $option{'tproto'});
    my_exit(1);
  }
}
#-------------------------------------------------------------------------------
#---------- start with program
# check if we got a name for a file that contents some filters
if ($option{'filter'} > 0) {
  if (! -f $option{'filter_file'}) {
    usage();
    printf(STDERR "error: file \"%s\" with filter not found.\n", $option{'filter_file'});
    my_exit(1);
  }
  # get filters
  if (read_filter($option{'filter_file'})) {
    my_exit(1);
  }
}

# correct splaytime
if ($option{'isplay'} < 0) {
  $option{'isplay'} = $default_values{'isplay'};
}

# print some info
if ($option{'quiet'} == 0) {

  # info about source
  print(STDOUT "info: running for source=\"$option{'iproto'}");
  if ($option{'iproto'} !~ /^stdin$/) {
    print(STDOUT ":$option{'local_addr'}");
  }
  if ($option{'local_port'} != 0) {
    print(STDOUT ":$option{'local_port'}");
  }

  # info about destination
  print(STDOUT "\"\n         and destination=\"$option{'oproto'}");
  if ($option{'oproto'} !~ /^stdout$/) {
    print(STDOUT ":$option{'remote_addr'}");
  }
  if ($option{'remote_port'} != 0) {
    print(STDOUT ":$option{'remote_port'}");
  }

  # info about tee destination
  if ($option{'tproto'} !~ /^$/) {
    print(STDOUT "\"\n         and destination=\"$option{'tproto'}");
    if ($option{'oproto'} !~ /^stdout$/) {
      print(STDOUT ":$option{'tee_addr'}");
    }
    if ($option{'tee_port'} != 0) {
      print(STDOUT ":$option{'tee_port'}");
    }
  }
  print(STDOUT "\"\n");


  # info for filter
  if ($option{'filter'} > 0) {
    print(STDOUT "         using ");
    if ($option{'filter'} == 1) {
      print(STDOUT "positive ");
    } else {
      print(STDOUT "negative ");
    }
    print(STDOUT "filter from file \"$option{'filter_file'}\"\n");
  }
  
  # info because no options are running program with stdin -> stdout
  if (($option{'iproto'} =~ /^stdin$/) && 
      ($option{'oproto'} =~ /^stdout$/)) {
    print(STDOUT "info: call program with option -h to get help message\n");
  }
}

# go in background if wanted
if ($option{'daemon'} == 1) {
  if ($option{'iproto'} =~ /^stdin$/) {
    print(STDERR "warning: reading from stdin is used .. so you can't daemonize this process\n" );
  } elsif ($option{'oproto'} =~ /^stdout$/) {
    print(STDERR "warning: writing to stdout is used .. so you can't daemonize this process\n" );
  } elsif ($option{'tproto'} =~ /^stdout$/) {
    print(STDERR "warning: writing to stdout is used by tee.. so you can't daemonize this process\n" );
  } else {
    daemonize();
  }
}

# open syslog
if ($option{'syslog'} == 1) {
  openlog($program, 'nowait,pid', 'user');
  $option{'syslog'} = 2;
}

my $src_is_not_open = 1;
my $retval = 0;
my $line = 0;

# our main loop
while (1) {
  local *SRC_FD;

  # re/open source if necessary
  if ($src_is_not_open != 0) {
    if ($option{'iproto'} =~ /^npipe$/) {
      $src_is_not_open = open_source_npipe();
    } elsif ($option{'iproto'} =~ /^tcp$/) {
      $src_is_not_open = open_source_tcp();
    } elsif ($option{'iproto'} =~ /^cmd$/) {
      $src_is_not_open = open_source_cmd();
    } elsif ($option{'iproto'} =~ /^stdin$/) {
      $file_desc{'src'} = *STDIN;
      $src_is_not_open = 0;
    } else {
      # protocol unknown for source
      if ($option{'syslog'} > 1) {
        syslog('error', 'protocol %s can not be used as source',
               $option{'iproto'});
      } else {
        printf(STDERR 'error: protocol %s can not be used as source\n',
               $option{'iproto'});
      }
      my_exit(1);
    }
  }

  # should we go sleep because source hasn't appeared yet?
  if (($src_is_not_open != 0) &&
      ($option{'isplay'} > 0)) {
    # go sleeping and sleep $option{'isplay'} seconds
    select(undef, undef, undef, $option{'isplay'});

    # try to open again
    next;
  }

  # process input and pipe it to output
  #print STDERR "main[$PID]: trying to read from src\n";
  *SRC_FD = $file_desc{'src'};
  while (defined($line = <SRC_FD>)) {
    # should we forward this?
    if (($option{'filter'} > 0) && reject_by_filter($line)) {
      # if reject_by_filter returns with 1
      # skip this line
      #print(STDERR "skip -->> $line");
      next;
    }

    # should we strip newline?
    if ($option{'strip_newline'} == 1) {
      chop($line);
    }

    # print debug
    if ($option{'print'} == 1) {
      if ($option{'syslog'} > 1) {
        syslog('notice', 'got[%d]: %s', $PID, chop($line));
      } else {
        printf(STDERR 'got[%d]: %s', $PID, $line);
      }
    }

    # send output to destination
    if ($option{'oproto'} =~ /^npipe$/) {
      # named pipe
      $retval = fwd_to_npipe('dst', 'remote_addr', $line);
    } elsif ($option{'oproto'} =~ /^tcp$/) {
      # tcp connection
      $retval = fwd_to_tcp('dst', 'remote_addr', 'remote_port', $line);
    } elsif ($option{'oproto'} =~ /^cmd$/) {
      # a normal working command
      $retval = fwd_to_cmd('dst', 'remote_addr', $line);
    } elsif ($option{'oproto'} =~ /^stdout$/) {
      # print to stdout
      print(STDOUT "$line");
      $retval = 0;
    }

    # send output to tee destination
    if ($option{'tproto'} !~ /^$/) {
      if ($option{'tproto'} =~ /^npipe$/) {
        # named pipe
        $retval = fwd_to_npipe('tee', 'tee_addr', $line);
      } elsif ($option{'tproto'} =~ /^tcp$/) {
        # tcp connection
        $retval = fwd_to_tcp('tee', 'tee_addr', 'tee_port', $line);
      } elsif ($option{'tproto'} =~ /^cmd$/) {
        # a normal working command
        $retval = fwd_to_cmd('tee', 'tee_addr', $line);
      } elsif ($option{'tproto'} =~ /^stdout$/) {
        # print to stdout
        print(STDOUT "$line");
        $retval = 0;
      }
    }

    # print debug
    if (($retval) && ($option{'print'} == 1)) {
      if ($option{'syslog'} > 1) {
        syslog('notice', 'skip[%d]: %s', $PID, chop($line));
      } else {
        printf(STDERR 'skip[%d]: %s', $PID, $line);
      }
    }
  }

  if ($option{'iproto'} =~ /^stdin$/) {
    # go sleeping $option{'isplay'} seconds or wait till somewhat
    # happens on stdin
    #select(fhbits('STDIN STDOUT'), undef, undef, $option{'isplay'});

    # exit because stdin may be closed
    my_exit(0);
  } else {
    # close source and try again
    #print STDERR "main[$PID]: closing src side\n";
    close($file_desc{'src'});
    $file_desc{'src'} = *MY_UNKNOWN_FD;
    $src_is_not_open = 1;
  }

  # check whether I'm a child (tcp proto)
  if ($PPID != $PID) {
    # yes then go away
    #print "main[$PID]: going away\n";
    my_exit(0);
  }

  # we try cmd only once
  if ($option{'iproto'} =~ /^cmd$/) {
    #print "main[$PID]: going away\n";
    my_exit(0);
  }
}

my_exit(0);
